package ims.hotcoref.symbols;

import ims.data.ISentence;
import ims.io.SentenceReader;
import ims.oracle.Configuration.TR;
import ims.trans.InstanceMapping;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.TransitionFactory;
import ims.util.DBO;
import ims.util.Util;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

public class SymbolTable implements Serializable {
	private static final long serialVersionUID = 6053605795146989988L;

	public static final String ZIP_ENTRY = "_symtab";
	
	public StringSymbolMapping
		forms,
		tags,
		feats,
		deprels,
		stags,
		ffeats,
		shapes;
	
	public StringSymbolMapping
		transitions;

	public ITransition[] transition2label;
	
	private boolean initialized=false;
	private boolean frozen=false;
	
	public SymbolTable(boolean initStags,boolean initTR,boolean initShapes){
		initTable(initStags,initTR,initShapes);
	}
	
	private void initTable(boolean initStags,boolean initTR,boolean initShapes){
		if(initialized)
			throw new UnsupportedOperationException("cant init twice");
		initialized=true;
		forms=new StringSymbolMapping("Form");
		tags=new StringSymbolMapping("POS");
		feats=new StringSymbolMapping("Feats");
		deprels=new StringSymbolMapping("Deprels");
		ffeats=new StringSymbolMapping("FullFeats");
		if(initTR)
			transitions=new StringSymbolMapping("Transitions");
		if(initStags)
			stags=new StringSymbolMapping("STags");
		shapes=new StringSymbolMapping("Shape");
	}
	
	public int registerSimple(SentenceReader<? extends ISentence> reader, int count, boolean labeled){
		return registerSimple(reader.iterator(),count,labeled);
	}
	public int registerSimple(Iterator<? extends ISentence> it, int count, boolean labeled){
		long start=System.currentTimeMillis();
		DBO.println("Registering simple types in symbol table:");
		DBO.printWithPrefix("Sent:  ");
		int erase=0;
		int sentCount=0;

		Set<String> leftLabels=null;
		Set<String> rightLabels=null;
		if(labeled){
			leftLabels=new TreeSet<String>();
			rightLabels=new TreeSet<String>();
		}
		
		while(it.hasNext()){
			ISentence s=it.next();
			if((sentCount-1)%500==0)
				erase=DBO.eraseAndPrint(erase,Integer.toString(sentCount));
			String[] form=s.getForms();
			String[] lemma=s.getTrainLemma();
			String[] pos=s.getParserTrainPOS();
			String[] feat=s.getFullTrainMorph();
			String[][] splitFeat=s.getTrainMorph();
			String[] lbl=s.getTrainLabels();
			String[] stag=s.getStag();
			int[] heads=s.getTrainHeads();
			for(int i=0;i<form.length;++i){
				forms.addSymbol(form[i]);
				forms.addSymbol(lemma[i]);
				tags.addSymbol(pos[i]);
				ffeats.addSymbol(feat[i]);
				for(String f:splitFeat[i])
					feats.addSymbol(f);
				deprels.addSymbol(lbl[i]);
				if(stags!=null)
					stags.addSymbol(stag[i]);
				if(shapes!=null)
					shapes.addSymbol(InstanceMapping.getShape(s, i));
			}
			if(labeled){
				for(int dep=1;dep<form.length;++dep){
					if(heads[dep]<dep){ //Right arc
						rightLabels.add(lbl[dep]);
					} else { //Left arc
						leftLabels.add(lbl[dep]);
					}
				}
			}
			if(sentCount>count)
				break;
			sentCount++;
		}
		erase=DBO.eraseAndPrint(erase,Integer.toString(sentCount));
		
		//Now register the transitions
		if(labeled){
			for(String lbl:leftLabels) //Leftarcs
				transitions.addSymbol(TR.LA.toString()+"_"+lbl);
			for(String lbl:rightLabels) //Rightarcs
				transitions.addSymbol(TR.RAS.toString()+"_"+lbl);
			transitions.addSymbols(TR.SH.toString()); //shift
		} else {
			transitions.addSymbols(TR.LA.toString(),TR.RAS.toString(),TR.SH.toString()); //LA, RA, SH
		}

		long time=System.currentTimeMillis()-start;
		DBO.printlnNoPrefix(",  time: "+Util.insertCommas(time));
		DBO.println();
		return sentCount;
	}
	
	public void addSwapTransition(){
		if(frozen)
			throw new Error("Already frozen symtab, can't add swap");
		transitions.addSymbol(TR.SW.toString());
	}
	
	public void addSBTransition(){
		if(frozen)
			throw new Error("Already frozen symtab, can't add swap");
		transitions.addSymbol(TR.SB.toString());
	}
	
//	private static final String NO_LABEL="<error>"; //<-- if we see this in the output, then there's some error... 
	
	public void freezeMappings(){
		if(frozen)
			throw new UnsupportedOperationException("cant freeze twice");
		
		DBO.println("Freezing symbol table");
		DBO.printlnNoPrefix("Types");
		
		if(this.forms!=null) 		{DBO.printlnNoPrefix(String.format("Forms:         %5d : %2d", forms.getItems(),forms.getBits())); 				forms.freeze();             }
		if(this.feats!=null) 		{DBO.printlnNoPrefix(String.format("Feats:         %5d : %2d", feats.getItems(),feats.getBits()));				feats.freeze();				}
		if(this.ffeats!=null) 		{DBO.printlnNoPrefix(String.format("FullFeats:     %5d : %2d", ffeats.getItems(),ffeats.getBits()));			ffeats.freeze();			}
		if(this.tags!=null) 		{DBO.printlnNoPrefix(String.format("Tags:          %5d : %2d", tags.getItems(),tags.getBits()));				tags.freeze();				}
		if(this.deprels!=null) 		{DBO.printlnNoPrefix(String.format("Deprels:       %5d : %2d", deprels.getItems(),deprels.getBits()));			deprels.freeze();			}
		if(this.transitions!=null) 	{DBO.printlnNoPrefix(String.format("Transitions:   %5d : %2d", transitions.getItems(),transitions.getBits()));	transitions.freeze();		}
		if(this.stags!=null)        {DBO.printlnNoPrefix(String.format("STags:         %5d : %2d", stags.getItems(),stags.getBits()));              stags.freeze();				}
		if(this.shapes!=null)       {DBO.printlnNoPrefix(String.format("Shape:         %5d : %2d", shapes.getItems(),shapes.getBits()));            shapes.freeze();			}
		//Establish a mapping from numeric transition -> transition object
		if(transitions!=null){
			transition2label=getITransionIndex();
		}
		
		frozen=true;
		DBO.println();
	}

	public ITransition[] getITransionIndex() {
		String[] list=transitions.getInverted();
		ITransition[] transition2label=new ITransition[list.length];
		TransitionFactory factory=new TransitionFactory();
		for(int i=0;i<list.length;++i){
			if(list[i]==null)
				continue;
			String op=list[i];
			transition2label[i]=factory.getTransition(op,i,deprels);
		}
		return transition2label;
	}

}
