package ims.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ims.data.Document;
import ims.data.ISentence;
import ims.util.AbstractSingleItemBufferedIterator;
import ims.util.IOUtils;

public class DocumentReader<T extends ISentence> implements Iterable<Document<T>> {

	private final Class<T> sentenceClass;
	private final File input;
	
	public DocumentReader(Class<T> sentenceClass,File input){
		this.sentenceClass=sentenceClass;
		this.input=input;
	}

	@Override
	public Iterator<Document<T>> iterator() {
		BufferedReader br=IOUtils.getBufferedReader(input);
		try {
			return new DocumentIterator(br);
		} catch (IOException e) {
			e.printStackTrace();
			throw new Error("!");
		}
	}
	
	public static <T extends ISentence> DocumentReader<T> getReader(Class<T> sentenceClass,File input){
		return new DocumentReader<T>(sentenceClass,input);
	}
	
	public Iterator<T> getSentenceIterator(){
		return new SentenceIterator(this.iterator());
	}

	class SentenceIterator implements Iterator<T> {

		Iterator<Document<T>> dIt;
		Iterator<T> sentIt; 
		
		public SentenceIterator(Iterator<Document<T>> it){
			this.dIt=it;
			if(it.hasNext()){
				Document<T> d=it.next();
				sentIt=d.sents.iterator();
			}
		}
		
		@Override
		public boolean hasNext() {
			return sentIt!=null && sentIt.hasNext();
		}

		@Override
		public T next() {
			T r=sentIt.next();
			if(!sentIt.hasNext()){
				if(dIt.hasNext()){
					Document<T> d=dIt.next();
					sentIt=d.sents.iterator();
				}
			}
			return r;
		}

		@Override
		public void remove() {
			throw new Error("not implemented");
		}
		
	}

	private static final Pattern STRIP_SENT_AND_DOC_ID_PATTERN=Pattern.compile("^(.*)\t(\\S+)\t(\\S+)$");
	class DocumentIterator extends AbstractSingleItemBufferedIterator<Document<T>> {

		private final BufferedReader br;
		
		public DocumentIterator(BufferedReader br) throws IOException{
			this.br=br;
			super.initSetNext(init());
		}
		
		@Override
		protected Document<T> readNext() throws IOException {
			List<T> sents=new ArrayList<>();
			List<String> cur=new ArrayList<>();
			List<String> sIDs=new ArrayList<>();
			String lDId=null;
			String lSId=null;
			String line;
			int sCount=0;
			while((line=br.readLine())!=null){
				if(line.trim().length()==0)
					break;
				Matcher m=STRIP_SENT_AND_DOC_ID_PATTERN.matcher(line);
				if(m.matches()){
					String l=m.group(1);
					String sId=m.group(2);
					String dId=m.group(3);
					if(lDId==null){ //First line altogether, no dId seen yet
						lDId=dId;
					} else if(!lDId.equals(dId)){
						throw new Error("Document ID mismatch: "+lDId+" vs. "+dId);
					}
					if(lSId==null){ //First line of sentence
						lSId=sId;
						sIDs.add(sId);
						cur.add(l);
					} else if(lSId.equals(sId)){ //Same sentence, go on
						cur.add(l);
					} else {
						T sent=TabbedReader.lines2Sentence(cur, sCount++, sentenceClass);
						sents.add(sent);
						cur.clear();
						cur.add(l);
						lSId=null;
					}
				} else {
					throw new Error("!");
				}
			}
			if(!cur.isEmpty()){
				T sent=TabbedReader.lines2Sentence(cur, sCount++, sentenceClass);
				sents.add(sent);				
			}
			if(sents.isEmpty())
				return null;
			Document.renormalizeHeadsAndIDsBySentences(sents);
			Document<T> d=new Document<>(sents,sIDs,lDId);
			return d;
		}

		@Override
		protected Document<T> init() throws IOException {
			return readNext();
		}
		
	}

}
