package ims.oracle;

import ims.data.Instance;
import ims.oracle.Configuration.TR;
import ims.trans.Options.TransitionSystem;
import ims.util.NonProjInfo;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class DeterministicArcStandardSwapOldOracle extends AbstractStaticOracle{

	protected DeterministicArcStandardSwapOldOracle(TR[] myTRs) {
		super(myTRs);
	}

	private InOrderOrder inOrder;
	
	@Override
	public void initSentence(int[] heads){
		inOrder=new InOrderOrder(InOrderOrder.getSequence(heads));
	}
	
	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels, boolean labeled) {
		Configuration c=new Configuration(heads);
		List<String> operations=new ArrayList<String>();
		boolean nonProj=!NonProjInfo.isProjectiveSentence(heads);
		InOrderOrder inOrder=nonProj?new InOrderOrder(InOrderOrder.getSequence(heads)):null;
		while(!(c.bufferSize()==0 && c.stackSize()==1)){
			final String op;
			if(DeterministicArcStandardOracle.canLeftArc(heads,c)){
				op=(labeled?TR.LA+"_"+labels[c.peekStack(1)]:TR.LA.toString());
				c.doLeftArc();
			} else if(DeterministicArcStandardOracle.canRightArc(heads,c)){
				op=(labeled?TR.RAS+"_"+labels[c.peekStack(0)]:TR.RAS.toString());
				c.doRightArcS();
			} else if(nonProj && canSwap(heads,c,inOrder)){
				op=TR.SW.toString();
				c.doSwap();
			} else {
				op=TR.SH.toString();
				c.doShift();
			}
			operations.add(op);
		}
		return operations;
	}
	
	static boolean canSwap(int[] heads, IConfiguration c,InOrderOrder inOrder) {
		if(!c.hasMinTwoStackElements())
			return false;
		int s1=c.peekStack(1);
		int s0=c.peekStack(0);
		
		return canSwap(s1,s0,inOrder);
		
//		int i=c.peekStack(1);
//		int j=c.peekStack(0);
//		if(inOrder.precedes(j, i))
//			return true;
//		else
//			return false;
	}
	
	static boolean canSwap(int s1,int s0,InOrderOrder inOrder){
		return inOrder.precedes(s0, s1);
	}

	@Override
	public boolean canDoNonProj() {
		return true;
	}


	public static final class InOrderOrder {
		
		final boolean[][] precedenceArray;
		
		public InOrderOrder(List<Integer> order){
			int size=order.size();
			precedenceArray=new boolean[size][size];
			for(int i=0;i<size;++i)
				for(int j=i+1;j<size;++j)
					precedenceArray[order.get(i)][order.get(j)]=true;
		}
		
		boolean precedes(int i,int j){
			return precedenceArray[i][j]; //If i < j
		}
		
		public static List<Integer> getSequence(int[] heads){
			Map<Integer,List<Integer>> left=new HashMap<Integer,List<Integer>>();
			Map<Integer,List<Integer>> right=new HashMap<Integer,List<Integer>>();
			for(int i=0;i<heads.length;++i){
				left.put(i,new ArrayList<Integer>());
				right.put(i,new ArrayList<Integer>());
			}
			for(int i=1;i<heads.length;++i){
				if(i<heads[i]){ //head is on the right, so we add to list of left deps
					left.get(heads[i]).add(i);
				} else { //head is on the left
					right.get(heads[i]).add(i);
				}
			}
			List<Integer> visited=new ArrayList<Integer>();
			descend(0,visited,left,right);
			return visited;
		}
		
		static void descend(Integer i,List<Integer> visited,Map<Integer,List<Integer>> left,Map<Integer,List<Integer>> right){
			for(Integer l:left.get(i))
				descend(l,visited,left,right);
			visited.add(i);
			for(Integer r:right.get(i))
				descend(r,visited,left,right);
		}
		
	}

	public static void main(String[] args){
		int[] heads=new int[]{-1,2,3,0,3,2,7,5,4,3};
		String[] tokens={"ROOT","A","hearing","is","scheduled","on","the","issue","today","."};
		List<Integer> sequence=InOrderOrder.getSequence(heads);
		for(Integer i:sequence){
			System.out.printf("%s < ",tokens[i]);
		}
		System.out.println();
		System.out.println();
		System.out.println();
		InOrderOrder order=new InOrderOrder(sequence);
		for(int[] pair:new int[][]{{2,5},{5,2},{7,4},{4,7},{3,5},{5,3}}){
			System.out.printf("%-10s < %-10s  ? %s\n",tokens[pair[0]],tokens[pair[1]],order.precedes(pair[0], pair[1]));
		}
		
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, Instance inst) {
		int[] heads=inst.head;
		if(tr==TR.SH){
			//Shift is OK if no other transition is OK.
			if(DeterministicArcStandardOracle.canLeftArc(heads, conf) || DeterministicArcStandardOracle.canRightArc(heads, conf) || canSwap(heads,conf,inOrder))
				return false;
			else
				return true;
		}
		switch(tr){
		case LA:  return DeterministicArcStandardOracle.canLeftArc(heads, conf);
		case RAS: return DeterministicArcStandardOracle.canRightArc(heads, conf);
		case SW:  return canSwap(heads,conf,inOrder);
		default:  throw new Error("not supported");
		}
	}
	
	@Override
	public TransitionSystem getTS() {
		return TransitionSystem.ArcSOldSwap;
	}
}
