package ims.tag;


import java.io.IOException;

import gnu.trove.map.hash.TIntIntHashMap;
import ims.data.ISentence;
import ims.hotcoref.symbols.TagSymbolTable;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntFactory;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.ParametersFloat.IUpdStruct;
import ims.perceptron.Regularizer;
import ims.tag.TagOptions.Decoder;
import ims.tag.features.AbstractFeatureSet;
import ims.tag.features.FeatureScorer;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.Pair;

public class Train {

	public static void main(String[] args) throws IOException, ClassNotFoundException{
		TagOptions options=new TagOptions(args);
		TagInstanceCreator tagInstanceCreator=new TagInstanceCreator(options.affixLength);
		TagSymbolTable symTab=new TagSymbolTable();
		SentenceReader<? extends ISentence> reader=ReaderWriterFactory.getReader(options.train, TagOptions.format);
		
		TagInstance[] instances=tagInstanceCreator.registerTypesAndCreateTrainInstances(reader.iterator(), symTab);
		
		AbstractFeatureSet fs=AbstractFeatureSet.getFS(options.fs,symTab);
		ParametersFloat params=new ParametersFloat(options.parameterSize,Regularizer.getRegularizer(options.regularizerType, 0.1, -1, -1));
		Long2IntInterface l2i=Long2IntFactory.getLong2Int(options.l2iImpl, options.parameterSize);
		FeatureScorer featureScorer=new FeatureScorer(fs, l2i, params.parameters, symTab);
		IDecoder decoder=getDecoder(options.decoder,featureScorer);
//		IDecoder decoder=fs.getDecoder(featureScorer);
		
		ArrayShuffler arrayShuffler=new ArrayShuffler(options.shuffleSeed);
		if(TagOptions.shuffleStrategy.shuffleFirstIteration)
			arrayShuffler.shuffle(instances);
		
		float upd=1;
		for(int iter=0;iter<options.iterations;++iter){
			int correct=0;
			double loss=0;
			int i=0;
			int erase=0;
			int failedUpdates=0;
			DBO.println("Iteration "+iter);
			for(TagInstance inst:instances){
				++i;
				int[] tags=decoder.decodeInstance(inst, symTab);
				if(!correctTags(tags,inst)){
					Pair<Double,Boolean> p=doUpdate(tags,inst,params,fs,upd,featureScorer,decoder);
					loss+=p.getLeft();
					if(p.getRight())
						++failedUpdates;
				} else {
					correct++;
				}
				++upd;
				if(i%333==0 || TagOptions.DEBUG){
					erase=DBO.eraseAndPrint(erase, String.format("I: %-5d  | C: %-5d   | l: %5.3f | f: %-5d (%5.2f%%)", i,correct,(loss/i),failedUpdates,(100.0*failedUpdates/i)));
					if(TagOptions.DEBUG){
						erase=0;
						DBO.println();
					}
						
				}
			}
			DBO.eraseAndPrint(erase, String.format("I: %-5d  | C: %-5d   | l: %5.3f | f: %-5d (%5.2f%%)", i,correct,(loss/i),failedUpdates,(100.0*failedUpdates/i)));
			DBO.println();
			if(options.testEveryIter && options.test!=null){
				l2i.freeze();
				Test.testDecoder(decoder, options.test, options.out, tagInstanceCreator, symTab);
				if(options.gold!=null)
					Test.eval(options.gold, options.out);
				l2i.unFreeze();
			}
			
			if(TagOptions.shuffleStrategy.shuffleBetweenIterations)
				arrayShuffler.shuffle(instances);
		}
		l2i.freeze();
		params.average(upd);
		
		if(options.model!=null){
			DBO.println("Saving model...");
			DBO.printlnNoPrefix("\n\nNOTE: Models do not save all hyperparameters of the decoders. Need to fix the code here!!!\n\n");
			TagModel tagModel=new TagModel(params.parameters, l2i, symTab, tagInstanceCreator,decoder);
			tagModel.saveModel(options.model);
			DBO.println("Done.");
		}
		
		if(options.test!=null && options.model!=null && !options.testEveryIter){
			Test.testModel(options.model, options.test, options.out);
		}
	}

	
	private static IDecoder getDecoder(Decoder decoder,FeatureScorer featureScorer) {
		switch(decoder){
		case Greedy:  return new GreedyLocalDecoder(featureScorer);
		case Collins: return new SecondOrderViterbiDecoder(featureScorer);
		case Sampler:
			ParametersFloat.MUTE_FAIL_UPDATES=true;
			return new SamplingDecoder(featureScorer);
		default: throw new Error("not implemented: "+decoder);
		}
	}


	static IFV<FV3> goldVector=new FV3();
	static IFV<FV3> predVector=new FV3();
	private static Pair<Double,Boolean> doUpdate(int[] tags, TagInstance inst, ParametersFloat params, AbstractFeatureSet fs, double upd, FeatureScorer featureScorer,IDecoder decoder) {
		double gScore=decoder.recoverFeaturesAndFillFV(inst.tags,inst,fs,goldVector,featureScorer);
		double pScore=decoder.recoverFeaturesAndFillFV(tags,inst,fs,predVector,featureScorer);
		double loss=getLoss(tags,inst.tags);
		TIntIntHashMap dv=goldVector.getDistVector(predVector);
		IUpdStruct u=params.getUpdStruct(pScore, gScore, loss, upd, dv);
		params.update(u);
		return new Pair<>(loss,u.failed());
	}

//	private static double recoverFeaturesAndFillFV(int[] tags, TagInstance inst, AbstractFeatureSet fs, IFV<FV3> fv,FeatureScorer featureScorer) {
//		TagContext tagContext=new TagContext(tags.length);
//		fv.clear();
//		double score=0.d;
//		for(int i=1;i<tags.length;++i){
//			score+=fs.recoverFeaturesAndFillFV(inst, tagContext, i, tags[i], featureScorer, fv);
//			tagContext.tags[i]=tags[i];
//		}
//		return score;
//	}

	private static double getLoss(int[] pred, int[] gold) {
		double l=0;
		for(int i=1;i<pred.length;++i)
			if(pred[i]!=gold[i])
				++l;
		return l;
	}

	private static boolean correctTags(int[] tags, TagInstance inst) {
		for(int i=1;i<inst.tags.length;++i){
			if(tags[i]!=inst.tags[i])
				return false;
		}
		return true;
	}
}
