package ims.trans;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import ims.data.Document;
import ims.data.ISentence;
import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.DocumentReader;
import ims.io.SentenceReader;
import ims.oracle.IOracle;
import ims.util.DBO;
import ims.util.StringShape;

public class InstanceMapping implements Serializable{
	private static final long serialVersionUID = -8824597142029396298L;

	final boolean useStag;
	final boolean useStack;
	final boolean useShape;
	
	public InstanceMapping(boolean useStag,boolean useStack,boolean useShape){
		this.useStag=useStag;
		this.useStack=useStack;
		this.useShape=useShape;
	}
	
	public Instance createTestInstance(ISentence s,SymbolTable symTab){
		return createTrainInstance(s,symTab,null,false);
	}
	
	public Instance createTrainInstance(ISentence s,SymbolTable symTab,IOracle oracle,boolean labeled){
		int[] seq=null;
		if(oracle!=null){
			seq=oracle.establishTransitionMappingAndGetCanonicalSequence(s, symTab.transitions, labeled);
		}
		Instance inst=new Instance(s.getLength(),seq,s.getSentId(),useStag,useStack,useShape);
		fillInstance(inst,s,symTab);
		return inst;
	}
	
	public List<Instance> createTrainInstances(SentenceReader<?> reader,SymbolTable symTab,int count,IOracle oracle,boolean labeled) {
		DBO.println("Creating instances");
		DBO.printWithPrefix("Sent:  ");
		List<Instance> instances=new ArrayList<Instance>(count);
		int erase=0;
		int i=0;
		for(ISentence s:reader){
			if(i>=count)
				break;
			if((i-1)%500==0)
				erase=DBO.eraseAndPrint(erase,Integer.toString(i));
			instances.add(createTrainInstance(s,symTab,oracle,labeled));
			++i;
		}
		erase=DBO.eraseAndPrint(erase,Integer.toString(i));
		DBO.println();
		return instances;
	}

	public Instance createTestInstance(Document<?> d, SymbolTable symTab) {
		List<Instance> instances=new ArrayList<>();
		for(ISentence s:d.sents){
			instances.add(createTestInstance(s,symTab));
		}
		Instance inst=Instance.mergeSentInstancesToOne(instances, useStag, useStack, useShape, -5,false,null, false);
		return inst;
	}
	
	public List<Instance> createTrainInstances(DocumentReader<?> reader,SymbolTable symTab,IOracle oracle,Options options){
		return createTrainInstances(reader,symTab,oracle,options.labeled,options.blockingSentenceBoundary);
	}
	
	public List<Instance> createTrainInstances(DocumentReader<?> reader,SymbolTable symTab,IOracle oracle,boolean labeled,boolean blockingSB){
		DBO.println("Creating instances");
		DBO.printWithPrefix("Doc:  ");
		List<Instance> instances=new ArrayList<Instance>();
		int erase=0;
		int i=0;
		for(Document<?> d:reader){
			List<Instance> sentInsts=new ArrayList<>();
			for(ISentence s:d.sents)
				sentInsts.add(createTrainInstance(s,symTab,oracle,labeled));
			erase=DBO.eraseAndPrint(erase,Integer.toString(i));
			Instance inst=Instance.mergeSentInstancesToOne(sentInsts,useStag,useStack,useShape,i,blockingSB,symTab,true);
			instances.add(inst);
			++i;
		}
		DBO.println();
		return instances;		
	}
	
	public void fillInstance(Instance instance, ISentence s, SymbolTable symTab) {
		String[] form=s.getForms();
		String[] lemma=s.getTrainLemma();
		String[] tag=s.getParserTrainPOS();
		String[][] splitFeat=s.getTrainMorph();
		int[] head=s.getTrainHeads();
		String[] lbl=s.getTrainLabels();
		for(int i=0;i<form.length;++i){
			instance.form[i]=symTab.forms.lookup(form[i]);
			instance.lemma[i]=symTab.forms.lookup(lemma[i]);
			instance.tag[i]=symTab.tags.lookup(tag[i]);
			if(splitFeat[i].length==0){
				instance.feat[i]=Instance.EMPTY_MORPH_ARR;
			} else {
				instance.feat[i]=new int[splitFeat[i].length];
				for(int p=0;p<splitFeat[i].length;++p)
					instance.feat[i][p]=symTab.feats.lookup(splitFeat[i][p]);
			}
			instance.head[i]=head[i];
			instance.lbl[i]=symTab.deprels.lookup(lbl[i]);
		}
		instance.head[0]=ISentence.ROOT_HEAD;
		if(useStag){
			String[] stags=s.getStag();
			for(int i=0;i<form.length;++i)
				fillStag(instance, symTab, stags[i], i);
		}
		if(useStack){
			int[] stackHead=s.getStackHead();
			String[] stackLbl=s.getStackLbl();
			for(int i=0;i<form.length;++i){
				instance.stackHead[i]=stackHead[i];
				instance.stackLbl[i]=symTab.deprels.lookup(stackLbl[i]);
			}
			instance.stackHasLeftDep[0]=false;
			instance.stackHasRightDep[0]=true;
			for(int h=1;h<form.length;++h){
				for(int l=h-1;l>0;--l){
					if(stackHead[l]==h){
						instance.stackHasLeftDep[h]=true;
						break;
					}
				}
				for(int r=h+1;r<form.length;++r){
					if(stackHead[r]==h){
						instance.stackHasRightDep[h]=true;
						break;
					}
				}
			}
		}
//		if(useShape){
		for(int i=0;i<form.length;++i){
			instance.shape[i]=symTab.shapes.lookup(getShape(s,i));
			instance.firstCharShape[i]=getFirstLastCharShape(s,i,true);
			instance.lastCharShape[i]=getFirstLastCharShape(s,i,false);
		}
//		}
	}


	private void fillStag(Instance instance, SymbolTable symTab, String stag, int i) {
		instance.stag[i]=symTab.stags.lookup(stag);
		int c=stag.length()-1;
		while(stag.charAt(c)!='/') //Look from beg
			--c;
		//now c is at the index of the first slash, so to the left is the label
		String stagLbl=stag.substring(0, c);
		++c;
		int headDir=stag.charAt(c)=='L'?Instance.STAG_LEFT:Instance.STAG_RIGHT;
		instance.stagLbl[i]=symTab.deprels.lookup(stagLbl);
		instance.stagHeadDir[i]=headDir;
		//Then it's the dependents
		c+=2;
		switch(stag.length()-c){
		case -1:  instance.stagDeps[i]=Instance.STAG_DEP_NONE; break;
		case 1:   instance.stagDeps[i]=(stag.charAt(c)=='L'?Instance.STAG_DEP_LEFT:Instance.STAG_DEP_RIGHT); break;
		default:  instance.stagDeps[i]=Instance.STAG_DEP_BOTH; break;
		}
	}
	
	public static String getShape(ISentence s,int index){
		if(index==0)
			return ISentence.ROOT;
		else
			return StringShape.reduceToShape(s.getForms()[index]);
	}
	
	public static int getFirstLastCharShape(ISentence s,int index,boolean first){
		if(index==0)
			return 6; //Shape for root
		final String form=s.getForms()[index];
		final char c;
		if(first)
			c=form.charAt(0);
		else
			c=form.charAt(form.length()-1);
		return StringShape.getCharShapeInt(c);
	}

}
