package ims.trans.doc;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.DocumentReader;
import ims.oracle.AbstractDynamicOracle;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.perceptron.Long2IntFactory;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.trans.AbstractTrainer;
import ims.trans.DynamicOracleTrainer;
import ims.trans.FeatureExtractor;
import ims.trans.GreedyTrainer;
import ims.trans.IInstanceCreator;
import ims.trans.InstanceMapping;
import ims.trans.LossFunction;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.trans.StaticListInstanceCreator;
import ims.trans.features.FeatureExtractorFactory;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFilterFactory;
import ims.trans.StaticOracleTrainer;
import ims.util.DBO;
import ims.util.Util;

import java.io.IOException;
import java.util.List;

public class Train {

	public static void main(String[] args) throws IOException, ClassNotFoundException{
		Options options=new Options(args);
		options.warnNoModel();
		
		if(options.testEveryIter && options.out==null){
			System.err.println("-testEvery but no -out given, stopping.");
			System.exit(1);
		}

		IOracle oracle;
		if(options.dynSentBoundaryOracle)
			oracle=OracleFactory.getDynSentOracle();
		else
			oracle=OracleFactory.getStaticOracle(TransitionSystem.ArcSNewSwap);
		
		SymbolTable symTab=new SymbolTable(options.useStag,true,options.useShape);

		InstanceMapping instanceMapping=new InstanceMapping(options.useStag,options.useStack,options.useShape);
		
		DocumentReader<?> reader=DocumentReader.getReader(Options.format.sentenceClass, options.train);
		
		List<Instance> instances = createInstances(options, reader, oracle, symTab,	instanceMapping);
		
		double avgTransSeqLen=ims.trans.Train.getAvgTransSeqLen(instances);
		DBO.printlnNoPrefix("Average transition sequence length (canonical):   "+String.format("%8.2f",avgTransSeqLen));

		LossFunction lossFunction=LossFunction.getLossFunction(Options.lossFunctionType);
		Long2IntInterface l2i=Long2IntFactory.getLong2Int(options);
		ParametersFloat parameters=ParametersFloat.createParameters(options);
		ITransitionFilter transitionFilter=TransitionFilterFactory.getDocumentTransitionFilter(symTab,options);
		FeatureExtractor featureExtractor=FeatureExtractorFactory.getExtractor(symTab,options);
		if(options.initBias)
			featureExtractor.initWeights(symTab, parameters, l2i);

		AbstractTrainer trainer;
		if(options.trainGreedy){
			options.beam=1;
			trainer=new GreedyTrainer(options, symTab, oracle, transitionFilter, featureExtractor, l2i, parameters, lossFunction, instanceMapping);
		} else {
			if(options.dynSentBoundaryOracle)
				trainer=new DynamicOracleTrainer(options,symTab,(AbstractDynamicOracle) oracle,transitionFilter,featureExtractor,l2i,parameters,lossFunction,instanceMapping,Options.goldBeam);
			else
				trainer=new StaticOracleTrainer(options, symTab, oracle, transitionFilter, featureExtractor, l2i, parameters, lossFunction, instanceMapping, Options.goldBeam);
		}
		
		List<String> modelInfo=ims.trans.Train.getModelInfo(options, oracle, trainer);
		ims.trans.Train.printModelInfo(modelInfo);
		
		IInstanceCreator instanceCreator=new StaticListInstanceCreator(instances);
		
		Model m=trainer.train(new Test(),instanceCreator);
		if(options.model!=null){
			DBO.println("Done training, saving model");
			ims.trans.Train.saveModel(m,options.model,modelInfo);
		} else {
			DBO.println("Skipping saving model (no model file provided)");
		}
		DBO.println("Done.");
		options.done();
	}
	
	private static List<Instance> createInstances(Options options, DocumentReader<?> reader, IOracle oracle, SymbolTable symTab, InstanceMapping instanceMapping) {
		int instCount=symTab.registerSimple(reader.getSentenceIterator(), options.count, options.labeled);
		symTab.addSwapTransition();
		symTab.addSBTransition();
		List<Instance> l=instanceMapping.createTrainInstances(reader, symTab, oracle, options);
		symTab.freezeMappings();
		
		if (false){ //Just debugging, print out first TR seq.
			Instance i=l.get(0);
			String[] s=new String[i.transSequence.length];
			
			for(int q=0;q<s.length;++q){
				s[q]=symTab.transition2label[i.transSequence[q]].getBasicTransitionLabel();
			}
			
			System.out.println(Util.joinStrings(",", s));
		}
		
		return l;
	}
}
