package ims.trans.util;

import ims.io.Format;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.trans.AbstractDecoder;
import ims.trans.Decoder;
import ims.trans.Model;
import ims.trans.Test;
import ims.util.Accuracy;
import ims.util.DBO;
import ims.util.DoubleSum;
import ims.util.Pair;
import ims.util.ZipUtil;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.ZipOutputStream;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;
import org.kohsuke.args4j.spi.StringArrayOptionHandler;

public class CombineModels {

	static enum ComboMode {
		Sum,
		NormalizedSum
	}
	
	static enum Average {
		None,
		Total,
		NonZeroTotal
	}
	
	@Option(name="-kahan",usage="Use Kahan summation")
	boolean kahan=false;
	
	@Option(name="-seed",usage="Seed model")
	File seedModel;
	
	@Option(name="-model",usage="Output model")
	File outModel;

	@Option(name="-models",usage="List of models to combine",handler = StringArrayOptionHandler.class)
	String[] models;
	
	@Option(name="-comboMode",usage="How to combine")
	ComboMode comboMode=ComboMode.NormalizedSum;
	
	@Option(name="-average",usage="Average weights scheme")
	Average average=Average.None;
	
	@Option(name="-test",usage="Test input file")
	File test;
	
	@Option(name="-eval",usage="Gold test file")
	File eval;
	
	@Option(name="-out",usage="Output test file")
	File out;
	
	@Option(name="-format",usage="Test file format")
	Format format=Format.CoNLL09;
	
	private Model seed;
	private double[] intermediateSumParams;
	private double[] nonZeroCountParams;
	private DoubleSum nonZeroSum;
		
	private void init() throws ClassNotFoundException, IOException{
		DBO.println("Options");
		DBO.printlnNoPrefix("ComboMode: "+comboMode);
		DBO.printlnNoPrefix("Average:   "+average);
		DBO.printlnNoPrefix("Kahan:     "+kahan);
		DBO.println();
		DBO.println("Loading seed model "+seedModel);
		seed=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, seedModel, Model.class);
		intermediateSumParams=new double[seed.parameters.length];
		nonZeroCountParams=new double[seed.parameters.length];
		nonZeroSum=DoubleSum.getDoubleSum(kahan);
	}
	
	private void aggregateSums() throws ClassNotFoundException, IOException {
		DBO.println("Going through models");
		for(String sf:models){
			File f=new File(sf);
			DBO.println("Loading "+f);
			Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, f, Model.class);
			DoubleSum ds;
			if(comboMode==ComboMode.NormalizedSum){
				ds=addToNormSum(m);
			} else {
				ds=addToSum(m);
			}
			DBO.printlnNoPrefix("Param vector size: " +m.parameters.length);
			DBO.printlnNoPrefix("Non-zero weight:   " +ds.addendCount());
			DBO.printlnNoPrefix("Avg weight:        " +ds.getAvg());
			DBO.printlnNoPrefix("STD weight:        " +ds.getStd());
			DBO.printlnNoPrefix("==============");
			nonZeroSum.addDouble(ds.addendCount());
		}
	}
	
	private void printStats() {
		double avgNonZero=nonZeroSum.getAvg();
		double nonZeroSTD=nonZeroSum.getStd();
		DBO.println("Average number of non-zero weights: "+String.format("%.1f", avgNonZero));
		DBO.println("STD on nonzero:                     "+String.format("%.1f", nonZeroSTD));
	}

	private void doAverage() {
		if(average==Average.None)
			return;
		DBO.println("Averaging...");
		if(average==Average.Total){
			double total=models.length;
			for(int i=0;i<intermediateSumParams.length;++i)
				intermediateSumParams[i]/=total;
		} else if(average==Average.NonZeroTotal){
			for(int i=0;i<intermediateSumParams.length;++i)
				if(nonZeroCountParams[i]>0)
					intermediateSumParams[i]/=nonZeroCountParams[i];
		}
	}

	private void floatAndSave() throws IOException {
		DBO.println("Creating new model...");
		DoubleSum ds=DoubleSum.getDoubleSum(kahan);
		float[] params=new float[intermediateSumParams.length];
		for(int i=0;i<intermediateSumParams.length;++i){
			if(intermediateSumParams[i]!=0){
				params[i]=(float) intermediateSumParams[i];
				ds.addDouble(intermediateSumParams[i]);
			}
		}
		DBO.printlnNoPrefix("New model parameters: "+params.length);
		DBO.printlnNoPrefix("Non-zero params:      "+ds.addendCount());
		DBO.printlnNoPrefix("Avg weight:           "+ds.getAvg());
		DBO.printlnNoPrefix("STD weight:           "+ds.getStd());
		DBO.println();
		
		Model out=new Model(seed.fe,seed.symTab,params,seed.l2i,seed.transitionFilter,seed.beamSize,seed.mergeStates,seed.labeled,seed.instanceMapping,seed.ts,seed.reversed,seed.sampleK,seed.stoppingCriterion);
		
		double nonZeroAvg=nonZeroSum.getAvg();
		double increaseFactor=ds.addendCount()/nonZeroSum.getAvg();
		DBO.println();
		DBO.println("Avg. non-zero weights: "+String.format("%.1f",nonZeroAvg));
		DBO.println("Out non-zero weights:  "+String.format("%.1f",ds.addendCount()));
		DBO.println("Increase factor:       "+String.format("%.3f",increaseFactor));
		DBO.println();
		
		DBO.println("Saving model...");
		ZipOutputStream zos=ZipUtil.openZOS(outModel);
		ZipUtil.writeObjectAsZipEntry(zos, Model.ZIP_FILE_ENTRY, out);
		zos.close();
		DBO.println("Done.");
	}
	
	private DoubleSum addToSum(Model m) {
		DoubleSum ds=DoubleSum.getDoubleSum(kahan);
		for(int i=0;i<m.parameters.length;++i){
			float f=m.parameters[i];
			if(f!=0){
				double d=f;
				intermediateSumParams[i]+=d;
				nonZeroCountParams[i]++;
				ds.addDouble(d);
			}
		}
		return ds;
	}

	private DoubleSum addToNormSum(Model m1) {
		DoubleSum ds=DoubleSum.getDoubleSum(kahan);
		for(float f:m1.parameters){
			if(f!=0){
				ds.addDouble(f);
			}
		}
		
		double std=ds.getStd();
		
		for(int i=0;i<m1.parameters.length;++i){
			if(m1.parameters[i]!=0){
				double d=m1.parameters[i];
				d/=std;
				intermediateSumParams[i]+=d;
				nonZeroCountParams[i]++;
			}
		}
		
		return ds;
	}
	
	private void testAndEval() throws ClassNotFoundException, IOException {
		if(test!=null && out!=null){
			DBO.println("==================");
			Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, outModel, Model.class);
			DBO.println("Testing model on "+test);
			SentenceReader<?> reader=ReaderWriterFactory.getReader(test, format.sentenceClass, m.reversed);
			SentenceWriter writer=ReaderWriterFactory.getWriter(out,m.reversed);
			AbstractDecoder d=new Decoder(m);
			Test.test(d, reader, writer);
		} else {
			return;
		}
		
		if(eval!=null){
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(out, eval, format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		} else {
			DBO.printlnNoPrefix("Skipping evaluation -- no -eval given");
		}
	}
	
	public static void main(String[] args) throws ClassNotFoundException, IOException{
		CombineModels combineModels=new CombineModels();
		CmdLineParser clp=new CmdLineParser(combineModels);
		try {
			clp.parseArgument(args);
		} catch (CmdLineException e) {
			e.printStackTrace();
			System.err.println("Failed to parse cmd line, aborting.");
			System.err.println(e.getMessage());
			System.err.println();
			System.err.println("Valid options are:");
			printUsage(System.err);
			System.exit(1);
		}
		
		combineModels.init();
		combineModels.aggregateSums();
		combineModels.printStats();
		combineModels.doAverage();
		combineModels.floatAndSave();
		combineModels.testAndEval();
	}

	public static void printUsage(OutputStream out){
		CmdLineParser clp=new CmdLineParser(new CombineModels());
		clp.printUsage(out);
	}
}
